# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'QMat Package'
copyright = '2024 PinT Community'
author = 'PinT Community'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = ''


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.mathjax',
    'autoapi.extension',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.extlinks',
    'nbsphinx',
    'sphinx_rtd_theme',
    'sphinx_gallery.load_style',
    'sphinxcontrib.video',
    'myst_parser',
    'sphinx_favicon',
]

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
]

favicons = [
    "android-chrome-192x192.png",
    "android-chrome-512x512.png",
    "apple-touch-icon.png",
    "favicon-16x16.png",
    "favicon-32x32.png",
    "favicon.ico",
]

add_module_names = False
autoapi_type = 'python'
autoapi_dirs = ['../qmat']
autoapi_file_patterns = ['*.py']
autoapi_options = ['members', 'undoc-members']
autoapi_python_class_content = 'both'
autoapi_add_toctree_entry = False

napoleon_use_param = False
napoleon_use_keyword = False
napoleon_use_ivar = True

extlinks = {'repo': ('https://github.com/Parallel-in-Time/qmat/tree/main/%s', None)}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
source_suffix = ['.rst', '.md']

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'default'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_logo = 'sdc_logo.png'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "collapse_navigation": "false"
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = [
    'custom.css',
]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'qmatdoc'


# -- Options for nbsphinx galleries
# Using _images/ is a hack to get relocated images which have been included in the pages
nbsphinx_thumbnails = {
    
}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Submodule for QDelta coefficients based on algebraic approach
"""
import numpy as np
import scipy.linalg as spl

from qmat.qdelta import QDeltaGenerator, register


@register
class PIC(QDeltaGenerator):
    """Picard approximation (zeros)"""
    aliases = ["Picard"]

    def getQDelta(self, k=None):
        return self.QDelta


@register
class Exact(QDeltaGenerator):
    """Takes Q (exact approximation)"""
    aliases = ["EXACT"]

    def getQDelta(self, k=None):
        return self.storeAndReturn(self.Q)


@register
class LU(QDeltaGenerator):
    """LU approximation from Weiser"""

    def getQDelta(self, k=None):
        _, _, U = spl.lu(self.Q.T)
        return self.storeAndReturn(U.T)


@register
class LU2(QDeltaGenerator):
    """LU approximation from Weiser multiplied by 2"""

    def getQDelta(self, k=None):
        _, _, U = spl.lu(self.Q.T)
        return self.storeAndReturn(2*U.T)


@register
class QPar(QDeltaGenerator):
    """Approximation using diagonal of Q"""
    aliases = ["Qpar", "Qdiag"]

    def getQDelta(self, k=None):
        return self.storeAndReturn(np.diag(np.diag(self.Q)))


@register
class GS(QDeltaGenerator):
    """Approximation using lower triangular part of Q"""
    aliases = ["GaussSeidel"]

    def getQDelta(self, k=None):
        return self.storeAndReturn(np.tril(self.Q))



import pytest
import numpy as np

from qmat.utils import getClasses, numericalOrder
from qmat.qcoeff import Q_GENERATORS
from qmat.nodes import NODE_TYPES, QUAD_TYPES

SCHEMES = getClasses(Q_GENERATORS)

def nStepsForTest(scheme, useEmbeddedWeights=False):
    try:
        nSteps = scheme.CONV_TEST_NSTEPS
    except AttributeError:
        expectedOrder = scheme.orderEmbedded if useEmbeddedWeights else scheme.order

        if expectedOrder == 1:
            nSteps = [64, 128, 256]
        elif expectedOrder == 2:
            nSteps = [32, 64, 128]
        elif expectedOrder == 3:
            nSteps = [16, 32, 64]
        elif expectedOrder in [4, 5]:
            nSteps = [8, 16, 32]
        elif expectedOrder in [6, 7]:
            nSteps = [4, 8, 16]
        elif expectedOrder in [8, 9]:
            nSteps = [2, 4, 8]
        else:
            nSteps = [1, 2, 4]  # default value (very high order methods)
    return nSteps

u0 = 1
lam = 1j
T = 2*np.pi



@pytest.mark.parametrize("useEmbeddedWeights", [True, False])
@pytest.mark.parametrize("scheme", SCHEMES.keys())
def testDahlquist(scheme, useEmbeddedWeights):
    gen = SCHEMES[scheme].getInstance()

    if useEmbeddedWeights:
        try:
            gen.weightsEmbedded
        except NotImplementedError:
            return None

    expectedOrder = gen.orderEmbedded if useEmbeddedWeights else gen.order
    nSteps = nStepsForTest(gen, useEmbeddedWeights)
    err = [gen.errorDahlquist(lam, u0, T, nS, useEmbeddedWeights=useEmbeddedWeights) for nS in nSteps]
    order, rmse = numericalOrder(nSteps, err)
    assert rmse < 0.02, f"rmse to high ({rmse}) for {scheme}"
    assert abs(order-expectedOrder) < 0.1, f"Expected order {expectedOrder:.2f}, but got {order:.2f} for {scheme}"


@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodesType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4])
def testDahlquistCollocation(nNodes, nodesType, quadType, useEmbeddedWeights=False):
    gen = SCHEMES["Collocation"](nNodes, nodesType, quadType)
    if useEmbeddedWeights:
        try:
            gen.weightsEmbedded
        except NotImplementedError:
            return None
    scheme = f"Collocation({nNodes}, {nodesType}, {quadType})"
    nSteps = nStepsForTest(gen, useEmbeddedWeights)
    tEnd = T
    err = [gen.errorDahlquist(lam, u0, tEnd, nS, useEmbeddedWeights=useEmbeddedWeights) for nS in nSteps]
    order, rmse = numericalOrder(nSteps, err)
    expectedOrder = gen.orderEmbedded if useEmbeddedWeights else gen.order

    assert rmse < 0.02, f"rmse to high ({rmse}) for {scheme} : {err}"
    if nNodes < 4:
        eps = 0.1
    else:
        eps = 0.25  # less constraining conditions for higher order
    assert abs(order-expectedOrder) < eps, f"Expected order {expectedOrder:.2f}, but got {order:.2f} for {scheme} {err}"

    if hasattr(gen, 'quadType'):
        assert gen.rightIsNode == (gen.quadType in ['LOBATTO', 'RADAU-RIGHT'])



import pytest
import numpy as np

from qmat.qdelta import QDELTA_GENERATORS
import qmat.qdelta.timestepping as module
from qmat.utils import getClasses, numericalOrder
from qmat.qcoeff.collocation import Collocation
from qmat.nodes import NODE_TYPES, QUAD_TYPES
from qmat.sdc import errorDahlquistSDC, getOrderSDC

SCHEMES = getClasses(QDELTA_GENERATORS, module="timestepping")


@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4, 5, 6])
def testBE(nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes = coll.nodes
    QDelta = module.BE(nodes).getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert np.allclose(QDelta.sum(axis=1), nodes), \
        "sum over the columns is not equal to nodes"


@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4, 5, 6])
def testFE(nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes = coll.nodes
    QDelta = module.FE(nodes).getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert not np.diag(QDelta).any(), \
        "QDelta has not zero diagonal"
    assert np.allclose(QDelta.sum(axis=1)[1:], np.cumsum(np.diff(coll.nodes))), \
        "sum over the columns is not equal to cumsum of node differences"

    _, dTau = module.FE(nodes).genCoeffs(dTau=True)
    assert type(dTau) == np.ndarray, \
        f"dTau is not np.ndarray but {type(dTau)}"
    assert dTau.ndim == 1, \
        f"dTau is not 1D : {dTau}"
    assert dTau.size == nNodes, \
        f"dTau has not the correct size : {dTau}"
    assert np.allclose(dTau, coll.nodes[0]), \
        "dTau is not equal to nodes[0]"


@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4, 5, 6])
def testTRAP(nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes = coll.nodes
    QDelta = module.TRAP(nodes).getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"

    QDeltaFE = module.FE(nodes).getQDelta()
    QDeltaBE = module.BE(nodes).getQDelta()
    assert np.allclose(QDelta, (QDeltaFE+QDeltaBE)/2), \
        "QDelta is not the mean of QDeltaBE and QDeltaFE"

    _, dTau = module.TRAP(nodes).genCoeffs(dTau=True)
    assert type(dTau) == np.ndarray, \
        f"dTau is not np.ndarray but {type(dTau)}"
    assert dTau.ndim == 1, \
        f"dTau is not 1D : {dTau}"
    assert dTau.size == nNodes, \
        f"dTau has not the correct size : {dTau}"
    assert np.allclose(dTau, coll.nodes[0]/2), \
        "dTau is not equal to node[0] divided by 2"


@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4, 5, 6])
def testBEPAR(nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes = coll.nodes
    QDelta = module.BEPAR(nodes).getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert np.allclose(QDelta, np.diag(np.diag(QDelta))), \
        "QDelta is not diagonal"
    assert np.allclose(QDelta, np.diag(coll.nodes)), \
        "QDelta diagonal is not equal to nodes"


@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4, 5, 6])
def testTRAPAR(nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes = coll.nodes
    QDelta = module.TRAPAR(nodes).getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert np.allclose(QDelta, np.diag(np.diag(QDelta))), \
        "QDelta is not diagonal"
    assert np.allclose(QDelta, np.diag(coll.nodes)/2), \
        "QDelta diagonal is not equal to nodes divided by 2"

    _, dTau = module.TRAPAR(nodes).genCoeffs(dTau=True)
    assert type(dTau) == np.ndarray, \
        f"dTau is not np.ndarray but {type(dTau)}"
    assert dTau.ndim == 1, \
        f"dTau is not 1D : {dTau}"
    assert dTau.size == nNodes, \
        f"dTau has not the correct size : {dTau}"
    assert np.allclose(dTau, coll.nodes/2), \
        "dTau is not equal to nodes divided by 2"


def nStepsForTest(order):
    nSteps = [1, 2, 4]  # default value (very high order methods)
    if order == 1:
        nSteps = [64, 128, 256]
    elif order == 2:
        nSteps = [32, 64, 128]
    elif order == 3:
        nSteps = [16, 32, 64]
    elif order in [4, 5]:
        nSteps = [8, 16, 32]
    elif order in [6, 7]:
        nSteps = [4, 8, 16]
    return nSteps

u0 = 1
lam = 1j
T = 2*np.pi

@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4])
@pytest.mark.parametrize("qDelta", SCHEMES.keys())
@pytest.mark.parametrize("nSweeps", [0, 1, 2, 3])
def testSDCConvergenceQUADRATURE(nSweeps, qDelta, nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes, weights, Q = coll.genCoeffs()
    QDelta = SCHEMES[qDelta](nodes).getQDelta()

    orderSDC = getOrderSDC(coll, nSweeps, qDelta, "QUADRATURE")
    nSteps = nStepsForTest(orderSDC)

    # Edge cases
    if qDelta == "BEPAR":
        if nSweeps == 3 and nNodes == 2 and nodeType == "LEGENDRE" and quadType == "RADAU-RIGHT":
            nSteps = nStepsForTest(orderSDC+1)
        if nSweeps == 3 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-RIGHT":
            nSteps = nStepsForTest(orderSDC+1)

    err = [
        errorDahlquistSDC(
            lam, u0, T, nS, nSweeps, Q, QDelta, weights)
        for nS in nSteps
        ]

    order, rmse = numericalOrder(nSteps, err)
    assert rmse < 0.04, f"rmse to high ({rmse})"
    assert abs(order-orderSDC) < 0.5, f"wrong numerical order ({order})"


@pytest.mark.parametrize("quadType", ["LOBATTO", "RADAU-RIGHT"])
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4])
@pytest.mark.parametrize("qDelta", SCHEMES.keys())
@pytest.mark.parametrize("nSweeps", [1, 2, 3, 4])
def testSDCConvergenceLASTNODE(nSweeps, qDelta, nNodes, nodeType, quadType):
    coll = Collocation(nNodes, nodeType, quadType)
    nodes, _, Q = coll.genCoeffs()
    QDelta = SCHEMES[qDelta](nodes).getQDelta()

    orderSDC = getOrderSDC(coll, nSweeps, qDelta, "LASTNODE")
    nSteps = nStepsForTest(orderSDC)

    # Edge cases
    if qDelta == "BEPAR":
        if nSweeps == 4 and nNodes == 3 and nodeType == "EQUID" and quadType == "LOBATTO":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-RIGHT":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType in "LEGENDRE":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType in "CHEBY-1" and quadType == "LOBATTO":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-1" and quadType == "RADAU-RIGHT":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-2" and quadType == "LOBATTO":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-2" and quadType == "RADAU-RIGHT":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-3":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-4":
            nSteps = nStepsForTest(orderSDC-1)
        if nSweeps == 4 and nNodes == 4:
            nSteps = nStepsForTest(orderSDC-1)


    err = [
        errorDahlquistSDC(
            lam, u0, T, nS, nSweeps, Q, QDelta, weights=None)
        for nS in nSteps
        ]

    # import matplotlib.pyplot as plt
    # plt.plot(T/np.array(nSteps), err)
    # import pdb; pdb.set_trace()

    order, rmse = numericalOrder(nSteps, err)
    assert rmse < 0.05, f"rmse to high ({rmse}), err={err}"
    assert abs(order-orderSDC) < 0.5, f"wrong numerical order ({order}), err={err}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Submodule to generate Q matrices based on Collocation
"""
import numpy as np

from qmat.qcoeff import QGenerator, register
from qmat.nodes import NodesGenerator
from qmat.lagrange import LagrangeApproximation

@register
class Collocation(QGenerator):
    aliases = ["coll"]

    DEFAULT_PARAMS = {
        "nNodes": 4,
        "nodeType": "LEGENDRE",
        "quadType": "RADAU-RIGHT",
        }
    
    def __init__(self, nNodes, nodeType, quadType, tLeft=0, tRight=1):
        self.nodeType, self.quadType = nodeType, quadType

        # Generate nodes between [0, 1]
        nodes = NodesGenerator(nodeType, quadType).getNodes(nNodes)
        nodes += 1
        nodes /= 2
        # Scale to [tLeft, tRight]
        nodes *= (tRight-tLeft)
        nodes += tLeft
        self.tLeft, self.tRight = tLeft, tRight
        # Rounding for safety
        np.round(nodes, 14, out=nodes) # TODO : check if necessary ...
        self._nodes = nodes

        # Lagrange approximation based on nodes
        approx = LagrangeApproximation(nodes)
        self._approx = approx

        # Compute Q (quadrature matrix) and weights
        Q = approx.getIntegrationMatrix([(tLeft, tau) for tau in nodes])
        weights = approx.getIntegrationMatrix([(tLeft, tRight)]).ravel()
        self._Q, self._weights = Q, weights

        # For convergence tests
        if nodes.size == 3 and nodeType in ["CHEBY-3", "CHEBY-4"]:
            self.CONV_TEST_NSTEPS = [32, 64, 128]  # high error constant

    @property
    def nodes(self): return self._nodes

    @property
    def Q(self): return self._Q

    @property
    def weights(self): return self._weights

    @property
    def S(self):
        nodes = self._nodes
        pInts = [(self.tLeft if i == 0 else nodes[i-1], nodes[i])
                 for i in range(nodes.shape[0])]
        return self._approx.getIntegrationMatrix(pInts)

    @property
    def hCoeffs(self):
        return self._approx.getInterpolationMatrix([1]).ravel()

    @property
    def order(self):
        M, nodeType, quadType = self.nodes.size, self.nodeType, self.quadType
        if nodeType != "LEGENDRE":
            if quadType in ["GAUSS", "LOBATTO"] \
                and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2"]:
                return M + (M % 2)  # why ? the node symmetry I guess ...
            return M
        else:
            quadType = self.quadType
            if quadType == "GAUSS":
                return 2*M
            elif quadType.startswith("RADAU"):
                return 2*M-1
            elif quadType == "LOBATTO":
                return 2*M-2



import pytest
import numpy as np

from qmat.qdelta import QDELTA_GENERATORS
import qmat.qdelta.min as module
from qmat.qcoeff.collocation import Collocation
from qmat.nodes import NODE_TYPES, QUAD_TYPES


def stiffSR(Q, QDelta):
    I = np.eye(Q.shape[0])
    M = I - np.linalg.solve(QDelta, Q)
    return max(abs(np.linalg.eigvals(M)))

# TODO : stiff spectral radius is not the same on different system ... to investigate
margin = 2

STIFF_SR = {
    "MIN": 0.43,
    "VDHS": 0.026,
    "MIN3": 0.0082,
    "MIN-SR-S": 0.00025,
    }
STIFF_PARAMS = {
    "nNodes": 4,
    "nodeType": "LEGENDRE",
    "quadType": "RADAU-RIGHT",
    }

@pytest.mark.parametrize("name", STIFF_SR.keys())
def testStiff4LegendreRadauRight(name):
    coll = Collocation(**STIFF_PARAMS)
    gen = QDELTA_GENERATORS[name](Q=coll.Q, **STIFF_PARAMS)

    QDelta = gen.getQDelta()
    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert np.allclose(QDelta, np.diag(np.diag(QDelta))), \
        "QDelta is not diagonal"

    sr = stiffSR(coll.Q, QDelta)
    assert sr < STIFF_SR[name] * margin, "spectral radius too high"


def nilpotencyNonStiff(Q, QDelta):
    nNodes = QDelta.shape[0]
    P = np.linalg.matrix_power(Q - QDelta, nNodes)
    return np.linalg.norm(P, ord=np.inf)

@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4, 5, 6, 7])
def testNonStiff(nNodes, nodeType, quadType):
    coll = Collocation(nNodes=nNodes, nodeType=nodeType, quadType=quadType)
    nodes, Q = coll.nodes, coll.Q

    gen = module.MIN_SR_NS(nodes)
    QDelta = gen.getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert np.allclose(QDelta, np.diag(np.diag(QDelta))), \
        "QDelta is not diagonal"
    assert nilpotencyNonStiff(Q, QDelta) < 1e-15 * margin, \
        "nilpotency measure is to high"


def nilpotencyStiff(Q, QDelta):
    if QDelta[0, 0] == 0:
        Q = Q[1:, 1:]
        QDelta = QDelta[1:, 1:]
    nNodes = QDelta.shape[0]
    I = np.eye(nNodes)
    P = np.linalg.matrix_power(I - np.linalg.solve(QDelta, Q), nNodes)
    return np.linalg.norm(P, ord=np.inf)

@pytest.mark.parametrize("quadType", QUAD_TYPES)
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4])
def testStiff(nNodes, nodeType, quadType):
    coll = Collocation(nNodes=nNodes, nodeType=nodeType, quadType=quadType)
    Q = coll.Q

    gen = module.MIN_SR_S(nNodes=nNodes, nodeType=nodeType, quadType=quadType)
    QDelta = gen.getQDelta()

    assert np.allclose(np.tril(QDelta), QDelta), \
        "QDelta is not lower triangular"
    assert np.allclose(QDelta, np.diag(np.diag(QDelta))), \
        "QDelta is not diagonal"
    assert nilpotencyStiff(Q, QDelta) < 1e-11 * margin, \
        "nilpotency measure is to high"


@pytest.mark.parametrize("quadType", ["GAUSS", "RADAU-RIGHT"])
@pytest.mark.parametrize("nodeType", NODE_TYPES)
@pytest.mark.parametrize("nNodes", [2, 3, 4])
def testFlex(nNodes, nodeType, quadType):
    coll = Collocation(nNodes=nNodes, nodeType=nodeType, quadType=quadType)
    Q = coll.Q

    gen = module.MIN_SR_FLEX(nNodes=nNodes, nodeType=nodeType, quadType=quadType)

    I = np.eye(nNodes)
    P = I
    for k in range(nNodes):
        P = (I - np.linalg.solve(gen.getQDelta(k+1), Q)) @ P

    assert np.allclose(np.tril(P), P), \
        "QDelta product is not lower triangular"
    assert np.allclose(P, np.diag(np.diag(P))), \
        "QDelta product is not diagonal"
    assert np.linalg.norm(P, ord=np.inf) < 1e-13 * margin, \
        "nilpotency measure is to high"

    genS = module.MIN_SR_S(nNodes=nNodes, nodeType=nodeType, quadType=quadType)

    QDeltaS = genS.getQDelta()
    QDeltaFlex = gen.getQDelta(nNodes+1)
    assert np.allclose(QDeltaS, QDeltaFlex), \
        "QDelta for k > nNodes is not equal to MIN-SR-S"

    QDelta0 = gen.getQDelta()
    QDelta1 = gen.getQDelta(1)
    assert np.allclose(QDelta0, QDelta1), \
        "default QDelta is not equal to k=1"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base module for Barycentric Lagrange Approximation
"""
import numpy as np
from scipy.special import roots_legendre


def computeFejerRule(n):
    """
    Compute a Fejer rule of the first kind, using DFT (Waldvogel 2006)
    Inspired from quadpy (https://github.com/nschloe/quadpy @Nico_Schlömer)

    Parameters
    ----------
    n : int
        Number of points for the quadrature rule.

    Returns
    -------
    nodes : np.1darray(n)
        The nodes of the quadrature rule
    weights : np.1darray(n)
        The weights of the quadrature rule.
    """
    # Initialize output variables
    n = int(n)
    nodes = np.empty(n, dtype=float)
    weights = np.empty(n, dtype=float)

    # Compute nodes
    theta = np.arange(1, n + 1, dtype=float)[-1::-1]
    theta *= 2
    theta -= 1
    theta *= np.pi / (2 * n)
    np.cos(theta, out=nodes)

    # Compute weights
    # -- Initial variables
    N = np.arange(1, n, 2)
    lN = len(N)
    m = n - lN
    K = np.arange(m)
    # -- Build v0
    v0 = np.concatenate([
        2 * np.exp(1j * np.pi * K / n) / (1 - 4 * K**2),
        np.zeros(lN + 1)])
    # -- Build v1 from v0
    v1 = np.empty(len(v0) - 1, dtype=complex)
    np.conjugate(v0[:0:-1], out=v1)
    v1 += v0[:-1]
    # -- Compute inverse fourier transform
    w = np.fft.ifft(v1)
    if max(w.imag) > 1.0e-15:
        raise ValueError(
            f'Max imaginary value to important for ifft: {max(w.imag)}')
    # -- Store weights
    weights[:] = w.real

    return nodes, weights


class LagrangeApproximation(object):
    r"""
    Class approximating any function on a given set of points using barycentric
    Lagrange interpolation.

    Let note :math:`(t_j)_{0\leq j<n}` the set of points, then any scalar
    function :math:`f` can be approximated by the barycentric formula :

    .. math::
        p(x) =
        \frac{\displaystyle \sum_{j=0}^{n-1}\frac{w_j}{x-x_j}f_j}
        {\displaystyle \sum_{j=0}^{n-1}\frac{w_j}{x-x_j}},

    where :math:`f_j=f(t_j)` and

    .. math::
        w_j = \frac{1}{\prod_{k\neq j}(x_j-x_k)}

    are the barycentric weights.
    The theory and implementation is inspired from [1]_.

    Parameters
    ----------
    points : list, tuple or np.1darray
        The given interpolation points, no specific scaling, but must be
        ordered in increasing order.
    weightComputation : str, optional
        Algorithm used to compute the barycentric weights. Can be :

        - 'FAST' : uses the analytic formula (unstable for large number of points)
        - 'STABLE' : uses logarithmic difference and scaling of the weights
        - 'CHEBFUN' : uses the same approach as in the chebfun package

        The default is 'AUTO' : it tries the 'FAST' algorithm, and if an
        overflow is detected, it switches to the 'STABLE' algorithm.
    scaleRef : str, optional
        Scaling used in the 'STABLE' algorithm for weight computation.
        Can be :

        - 'ZERO' : scaling based on the weight for the value closest to :math:`t=0`.
        - 'MAX' : scaling based on the maximum weight value.

        The default is 'MAX'.

    Attributes
    ----------
    points : np.1darray
        The interpolating points
    weights : np.1darray
        The associated barycentric weights
    n : int (property)
        The number of points

    .. [1] Berrut, J. P., & Trefethen, L. N. (2004).
           "Barycentric Lagrange interpolation." SIAM review, 46(3), 501-517.
    """

    def __init__(self, points, weightComputation='AUTO', scaleRef='MAX'):
        points = np.asarray(points).ravel()
        assert np.unique(points).size == points.size, "distinct interpolation points are required"

        diffs = points[:, None] - points[None, :]
        diffs[np.diag_indices_from(diffs)] = 1

        def analytic(diffs):
            # Fast implementation (unstable for large number of points)
            invProd = np.prod(diffs, axis=1)
            invProd **= -1
            return invProd

        def logScale(diffs):
            # Implementation using logarithmic difference and scaling
            sign = np.sign(diffs).prod(axis=1)
            wLog = -np.log(np.abs(diffs)).sum(axis=1)
            if scaleRef == 'ZERO':
                wScale = wLog[np.argmin(np.abs(points))]
            elif scaleRef == 'MAX':
                wScale = wLog.max()
            else:
                raise NotImplementedError(f'scaleRef={scaleRef}')
            invProd = np.exp(wLog - wScale)
            invProd *= sign
            return invProd

        def chebfun(diffs):
            # Implementation used in chebfun
            diffs *= 4 / (points.max() - points.min())
            sign = np.sign(diffs).prod(axis=1)
            vv = np.exp(np.log(np.abs(diffs)).sum(axis=1))
            invProd = (sign * vv)
            invProd **= -1
            invProd /= np.linalg.norm(invProd, np.inf)
            return invProd

        if weightComputation == 'AUTO':
            with np.errstate(divide='raise', over='ignore'):
                try:
                    invProd = analytic(diffs)
                except FloatingPointError:
                    invProd = logScale(diffs)
        elif weightComputation == 'FAST':
            invProd = analytic(diffs)
        elif weightComputation == 'STABLE':
            invProd = logScale(diffs)
        elif weightComputation == 'CHEBFUN':
            invProd = chebfun(diffs)
        else:
            raise NotImplementedError(
                f'weightComputation={weightComputation}')
        weights = invProd
        weights /= np.max(np.abs(weights))

        # Store attributes
        self.points = points
        self.weights = weights
        self.weightComputation = weightComputation


    @property
    def n(self):
        """The number of points"""
        return self.points.size


    def getInterpolationMatrix(self, times):
        r"""
        Compute the interpolation matrix for a given set of discrete "time"
        points.

        For instance, if we note :math:`\vec{f}` the vector containing the
        :math:`f_j=f(t_j)` values, and :math:`(\tau_m)_{0\leq m<M}`
        the "time" points where to interpolate.
        Then :math:`I[\vec{f}]`, the vector containing the interpolated
        :math:`f(\tau_m)` values, can be obtained using :

        .. math::
            I[\vec{f}] = P_{Inter} \vec{f},

        where :math:`P_{Inter}` is the interpolation matrix returned by this
        method.

        Parameters
        ----------
        times : list-like or np.1darray
            The discrete "time" points where to interpolate the function.

        Returns
        -------
        PInter : np.2darray(M, n)
            The interpolation matrix, with :math:`M` rows (size of the **times**
            parameter) and :math:`n` columns.

        """
        # Compute difference between times and Lagrange points
        times = np.asarray(times)
        with np.errstate(divide='ignore'):
            iDiff = 1 / (times[:, None] - self.points[None, :])

        # Find evaluated positions that coincide with one Lagrange point
        concom = (iDiff == np.inf) | (iDiff == -np.inf)
        i, j = np.where(concom)

        # Replace iDiff by on on those lines to get a simple copy of the value
        iDiff[i, :] = concom[i, :]

        # Compute interpolation matrix using weights
        PInter = iDiff * self.weights
        PInter /= PInter.sum(axis=-1)[:, None]

        return PInter


    def getIntegrationMatrix(self, intervals, numQuad='LEGENDRE_NUMPY'):
        r"""
        Compute the integration matrix for a given set of intervals.

        For instance, if we note :math:`\vec{f}` the vector containing the
        :math:`f_j=f(t_j)` values, and
        :math:`(\tau_{m,left}, \tau_{m,right})_{0\leq m<M}` the different
        interval where the function should be integrated using the barycentric
        interpolant polynomial.
        Then :math:`\Delta[\vec{f}]`, the vector containing the approximations
        of

        .. math::
            \int_{\tau_{m,left}}^{\tau_{m,right}} f(t)dt,

        can be obtained using :

        .. math::
            \Delta[\vec{f}] = P_{Integ} \vec{f},

        where :math:`P_{Integ}` is the interpolation matrix returned by this
        method.

        Parameters
        ----------
        intervals : list of pairs
            A list of all integration intervals.
        numQuad : str, optional
            Quadrature rule used to integrate the interpolant barycentric
            polynomial. Can be :

            - 'LEGENDRE_NUMPY' : Gauss-Legendre rule from Numpy
            - 'LEGENDRE_SCIPY' : Gauss-Legendre rule from Scipy
            - 'FEJER' : internaly implemented Fejer-I rule

            The default is 'LEGENDRE_NUMPY'.

        Returns
        -------
        PInter : np.2darray(M, n)
            The integration matrix, with :math:`M` rows (number of intervals)
            and :math:`n` columns.
        """
        if numQuad == 'LEGENDRE_NUMPY':
            # Legendre gauss rule, integrate exactly polynomials of deg. (2n-1)
            iNodes, iWeights = np.polynomial.legendre.leggauss((self.n + 1) // 2)
        elif numQuad == 'LEGENDRE_SCIPY':
            # Using Legendre scipy implementation
            iNodes, iWeights = roots_legendre((self.n + 1) // 2)
        elif numQuad == 'FEJER':
            # Fejer-I rule, integrate exactly polynomial of deg. n-1
            iNodes, iWeights = computeFejerRule(self.n - ((self.n + 1) % 2))
        else:
            raise NotImplementedError(f'numQuad={numQuad}')

        # Compute quadrature nodes for each interval
        intervals = np.array(intervals)
        aj, bj = intervals[:, 0][:, None], intervals[:, 1][:, None]
        tau, omega = iNodes[None, :], iWeights[None, :]
        tEval = (bj - aj) / 2 * tau + (bj + aj) / 2

        # Compute the integrand function on nodes
        integrand = self.getInterpolationMatrix(tEval.ravel()).T.reshape(
            (-1,) + tEval.shape)

        # Apply quadrature rule to integrate
        integrand *= omega
        integrand *= (bj - aj) / 2
        PInter = integrand.sum(axis=-1).T

        return PInter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Submodule for QDelta coefficients based on minimization approaches
"""
import warnings
import numpy as np
import scipy.optimize as spo

from qmat.qdelta import QDeltaGenerator, register
import qmat.qdelta.mincoeffs as tables
from qmat.qcoeff.collocation import Collocation


@register
class MIN(QDeltaGenerator):
    aliases = ["MIN-Speck"]

    def rho(self, x):
        M = self.QDelta.shape[0]
        return max(abs(
            np.linalg.eigvals(np.eye(M) - np.diag(x).dot(self.Q))))

    def getQDelta(self, k=None):
        x0 = 10 * np.ones(self.Q.shape[0])
        d = spo.minimize(self.rho, x0, method='Nelder-Mead')
        np.copyto(self.QDelta, np.linalg.inv(np.diag(d.x)))
        return self.QDelta


class FromTable(QDeltaGenerator):

    def __init__(self, nNodes, nodeType, quadType, **kwargs):
        self.nNodes = nNodes
        self.nodeType = nodeType
        self.quadType = quadType
        self.QDelta = np.zeros((nNodes, nNodes), dtype=float)

    def getQDelta(self, k=None):
        name = self.__class__.__name__
        try:
            table = getattr(tables, name)
            coeffs = table[self.nodeType][self.quadType][self.nNodes]
        except KeyError:
            raise NotImplementedError(
                "no {name} MIN coefficients for"
                f"{self.nNodes} {self.nodeType} {self.quadType} nodes")
        np.copyto(self.QDelta, np.diag(coeffs))
        return self.QDelta

    def check(cls):
        try:
            getattr(tables, cls.__name__)
        except AttributeError:
            raise AttributeError(
                f"no MIN coefficients table found for {cls.__name__}"
                " in qmat.qdelta.mincoeffs")
        return cls

def registerTable(cls:FromTable)->FromTable:
    return register(FromTable.check(cls))


@registerTable
class MIN3(FromTable):
    """
    These values have been obtained using Indie Solver, a commercial solver for
    black-box optimization which aggregates several state-of-the-art optimization
    methods (free academic subscription plan).
    Objective function :
        sum over 17^2 values of lamdt, real and imaginary
    (WORKS SURPRISINGLY WELL!)
    """
    aliases = ["Magic_Numbers"]


@registerTable
class MIN_VDHS(FromTable):
    aliases = ["VDHS"]


@register
class MIN_SR_NS(QDeltaGenerator):
    aliases = ["MIN-SR-NS", "MIN_GT"]

    def __init__(self, nodes, **kwargs):
        self.nodes = np.asarray(nodes)
        M = self.nodes.size
        self.QDelta = np.zeros((M, M), dtype=float)

    def getQDelta(self, k=None):
        np.copyto(self.QDelta, np.diag(self.nodes)/self.nodes.size)
        return self.QDelta


@register
class MIN_SR_S(QDeltaGenerator):
    aliases = ["MIN-SR-S"]

    def __init__(self, nNodes, nodeType, quadType, **kwargs):
        self.nNodes = nNodes
        self.nodeType = nodeType
        self.quadType = quadType
        self.QDelta = np.zeros((nNodes, nNodes), dtype=float)
        self.coll = Collocation(nNodes, nodeType, quadType)

    def computeCoeffs(self, M, a=None, b=None):
        """
        Compute diagonal coefficients for a given number of nodes M.
        If `a` and `b` are given, then it uses as initial guess:

        >>> a * nodes**b / M

        If `a` is not given, then do not care about `b` and uses as initial guess:

        >>> nodes / M

        Parameters
        ----------
        M : int
            Number of collocation nodes.
        a : float, optional
            `a` coefficient for the initial guess.
        b : float, optional
            `b` coefficient for the initial guess.

        Returns
        -------
        coeffs : array
            The diagonal coefficients.
        nodes : array
            The nodes associated to the current coefficients.
        """
        nodeType, quadType = self.nodeType, self.quadType
        if M == self.nNodes:
            collM = self.coll
        else:
            collM = Collocation(nNodes=M, nodeType=nodeType, quadType=quadType)
        QM, nodesM = collM.Q, collM.nodes

        if quadType in ['LOBATTO', 'RADAU-LEFT']:
            QM = QM[1:, 1:]
            nodesM = nodesM[1:]
        nCoeffs = len(nodesM)

        if nCoeffs == 1:
            coeffs = np.diag(QM)

        else:

            def nilpotency(coeffs):
                """Function verifying the nilpotency from given coefficients"""
                coeffs = np.asarray(coeffs)
                kMats = [(1 - z) * np.eye(nCoeffs) + z * np.diag(1 / coeffs) @ QM for z in nodesM]
                vals = [np.linalg.det(K) - 1 for K in kMats]
                return np.array(vals)

            if a is None:
                coeffs0 = nodesM / M
            else:
                coeffs0 = a * nodesM**b / M

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                coeffs = spo.fsolve(nilpotency, coeffs0, xtol=1e-15)

        # Handle first node equal to zero
        if quadType in ['LOBATTO', 'RADAU-LEFT']:
            coeffs = np.asarray([0.0] + list(coeffs))
            nodesM = np.asarray([0.0] + list(nodesM))

        return coeffs, nodesM

    @staticmethod
    def fit(coeffs, nodes):
        """Function fitting given coefficients to a power law"""

        def lawDiff(ab):
            a, b = ab
            return np.linalg.norm(a * nodes**b - coeffs)

        sol = spo.minimize(lawDiff, [1.0, 1.0], method="nelder-mead")
        return sol.x

    def getQDelta(self, k=None):
        # Compute coefficients incrementally
        a, b = None, None
        m0 = 2 if self.quadType in ['LOBATTO', 'RADAU-LEFT'] else 1
        for m in range(m0, self.nNodes + 1):
            coeffs, nodes = self.computeCoeffs(m, a, b)
            if m > 1:
                a, b = self.fit(coeffs * m, nodes)

        np.copyto(self.QDelta, np.diag(coeffs))
        return self.QDelta


@register
class MIN_SR_FLEX(MIN_SR_S):
    aliases = ["MIN-SR-FLEX"]

    def getQDelta(self, k=None):
        if k is None:
            k = 1
        if k < 1:
            raise ValueError(f"k must be greater than 0 ({k})")
        nodes = self.coll.nodes
        if k <= nodes.size:
            np.copyto(self.QDelta, np.diag(nodes)/k)
        else:
            try:
                self.QDelta_MIN_SR_S
            except AttributeError:
                self.QDelta_MIN_SR_S = super().getQDelta()
            np.copyto(self.QDelta, self.QDelta_MIN_SR_S)
        return self.QDelta



import pytest
import numpy as np

from qmat.qcoeff import Q_GENERATORS, genQCoeffs
from qmat.utils import getClasses

GENERATORS = getClasses(Q_GENERATORS)


@pytest.mark.parametrize("name", GENERATORS.keys())
def testGeneration(name):
    gen = GENERATORS[name].getInstance()
    n1, w1, Q1 = gen.nodes, gen.weights, gen.Q

    assert type(n1) == np.ndarray, \
        f"nodes for {name} are not np.ndarray but {type(n1)}"
    assert type(w1) == np.ndarray, \
        f"weights for {name} are not np.ndarray but {type(w1)}"
    assert type(Q1) == np.ndarray, \
        f"Q for {name} are not np.ndarray but {type(Q1)}"

    assert n1.ndim == 1, f"nodes for {name} are not 1D : {n1}"
    assert w1.ndim == 1, f"weights for {name} are not 1D : {n1}"
    assert n1.size == w1.size, \
        f"nodes and weights for {name} don't have the same size : {n1} / {w1}"
    assert Q1.shape == (n1.size, n1.size), \
        f"Q for {name} has inconsistent shape : {Q1.shape}"
    assert gen.nNodes == n1.size, \
        f"nNodes property from {name} is not equal to node size !"

    try:
        n2, w2, Q2 = genQCoeffs(name)
    except TypeError:
        n2, w2, Q2 = genQCoeffs(name, **GENERATORS[name].DEFAULT_PARAMS)
    assert np.allclose(n1, n2), \
        f"OOP nodes {n1} and PP nodes {n2} are not equals for {name}"
    assert np.allclose(w1, w2), \
        f"OOP weights {w1} and PP weights {w2} are not equals for {name}"
    assert np.allclose(Q1, Q2), \
        f"OOP Q matrix {Q1} and PP Q matrix {Q2} are not equals for {name}"


@pytest.mark.parametrize("name", GENERATORS.keys())
def testAdditionalCoeffs(name):
    gen = GENERATORS[name].getInstance()
    nodes, S1, h1 = gen.nodes, gen.S, gen.hCoeffs

    assert type(S1) == np.ndarray, \
        f"S for {name} are not np.ndarray but {type(S1)}"
    assert type(h1) == np.ndarray, \
        f"hCoeffs for {name} are not np.ndarray but {type(h1)}"

    assert S1.shape == (nodes.size, nodes.size), \
        f"S for {name} has inconsistent shape : {S1.shape}"
    assert h1.ndim == 1, f"hCoeffs for {name} are not 1D : {h1}"
    assert h1.size == nodes.size, \
        f"hCoeffs for {name} has inconsistent size : {h1.size}"

    try:
        _, _, _, S2, h2 = genQCoeffs(name, withS=True, hCoeffs=True)
    except TypeError:
        _, _, _, S2, h2 = genQCoeffs(name, withS=True, hCoeffs=True,
                                **GENERATORS[name].DEFAULT_PARAMS)
    assert np.allclose(S1, S2), \
        f"OOP S matrix {S1} and PP S matrix {S2} are not equals for {name}"
    assert np.allclose(h1, h2), \
        f"OOP hCoeffs {h1} and PP hCoeffs {h2} are not equals for {name}"


    try:
        try:
            _, b, _  = genQCoeffs(name, embedded=True)
        except TypeError:
            _, b, _  = genQCoeffs(name, embedded=True, **GENERATORS[name].DEFAULT_PARAMS)

        assert type(b) == np.ndarray
        assert b.ndim == 2
    except NotImplementedError:
        pass


@pytest.mark.parametrize("name", GENERATORS.keys())
def testS(name):
    gen = GENERATORS[name].getInstance()
    Q, S, T, Tinv = gen.Q, gen.S, gen.T, gen.Tinv

    Q2S = T @ Q
    assert np.allclose(Q2S, S), f"Q transformed to S is not equal to S for {name}"

    Q2S2Q = Tinv @ Q2S
    assert np.allclose(Q2S2Q, Q), \
        f"Q transformed to S then back to Q is not equal to Q for {name}"



import pytest
import numpy as np

from qmat.nodes import NodesGenerator, NODE_TYPES

def chebyNodes(kind, n):
    i = np.arange(n, dtype=float) + 1
    i = i[-1::-1]
    if kind == 1:
        nodes = np.cos((i - 0.5) / n * np.pi)
    elif kind == 2:
        nodes = np.cos(i / (n + 1) * np.pi)
    elif kind == 3:
        nodes = np.cos((i - 0.5) / (n + 0.5) * np.pi)
    elif kind == 4:
        nodes = np.cos(i / (n + 0.5) * np.pi)
    return tuple(nodes)

def equidNodes(n):
    return tuple(np.linspace(-1, 1, num=n + 2)[1:-1])

REF_NODES = {
    'LEGENDRE': {
        2: (-1 / 3**0.5, 1 / 3**0.5),
        3: (-((3 / 5) ** 0.5), 0, (3 / 5) ** 0.5),
        4: (
            -((3 / 7 + 2 / 7 * (6 / 5) ** 0.5) ** 0.5),
            -((3 / 7 - 2 / 7 * (6 / 5) ** 0.5) ** 0.5),
            (3 / 7 - 2 / 7 * (6 / 5) ** 0.5) ** 0.5,
            (3 / 7 + 2 / 7 * (6 / 5) ** 0.5) ** 0.5,
        ),
        5: (
            -1 / 3 * (5 + 2 * (10 / 7) ** 0.5) ** 0.5,
            -1 / 3 * (5 - 2 * (10 / 7) ** 0.5) ** 0.5,
            0,
            1 / 3 * (5 - 2 * (10 / 7) ** 0.5) ** 0.5,
            1 / 3 * (5 + 2 * (10 / 7) ** 0.5) ** 0.5,
        ),
    }
}
nTests = list(REF_NODES['LEGENDRE'].keys())
for kind in [1, 2, 3, 4]:
    REF_NODES[f'CHEBY-{kind}'] = {n: chebyNodes(kind, n) for n in nTests}
REF_NODES['EQUID'] = {n: equidNodes(n) for n in nTests}


@pytest.mark.parametrize("nodeType", REF_NODES.keys())
def testGauss(nodeType):
    gen = NodesGenerator(nodeType=nodeType, quadType='GAUSS')
    ref = REF_NODES[nodeType]
    for n, nodes in ref.items():
        assert np.allclose(nodes, gen.getNodes(n)), f"difference with reference nodes (n={n})"


@pytest.mark.parametrize("quadType", ["LOBATTO", "RADAU-RIGHT", "RADAU-LEFT"])
@pytest.mark.parametrize("nodeType", NODE_TYPES)
def testNonGauss(nodeType, quadType):
    gen = NodesGenerator(nodeType=nodeType, quadType=quadType)
    for n in nTests:
        nodes = gen.getNodes(n)
        if quadType in ["LOBATTO", "RADAU-RIGHT"]:
            assert nodes[-1] == 1, f"right node not equal to 1 (n={n})"
        if quadType in ["LOBATTO", "RADAU-LEFT"]:
            assert nodes[0] == -1, f"left node not equal to -1 (n={n})"
        assert np.allclose(np.sort(nodes), nodes), f"nodes not ordered (n={n})"


@pytest.mark.parametrize("quadType", ["GAUSS", "LOBATTO", "RADAU-RIGHT", "RADAU-LEFT"])
@pytest.mark.parametrize("nodeType", ["LEGENDRE", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"])
def testAsymptotic(nodeType, quadType):
    gen = NodesGenerator(nodeType=nodeType, quadType=quadType)
    unif = NodesGenerator(nodeType="EQUID", quadType=quadType)
    n = 1000
    nodes = gen.getNodes(n)
    s = (unif.getNodes(n) + 1)/2
    limit = -np.cos(s*np.pi)
    assert np.max(np.abs(limit-nodes)) < 0.01



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base module for nodes generation
"""
import numpy as np
from scipy.linalg import eigh_tridiagonal

NODE_TYPES = ['EQUID', 'LEGENDRE', 'CHEBY-1', 'CHEBY-2', 'CHEBY-3', 'CHEBY-4']
"""Available types (distributions) for nodes"""

QUAD_TYPES = ['GAUSS', 'RADAU-LEFT', 'RADAU-RIGHT', 'LOBATTO']
"""Available quadrature type for each node distribution"""

class NodesGenerator(object):
    """
    Class that can be used to generate generic distribution of nodes derived
    from Gauss quadrature rule.
    Its implementation is fully inspired from a `book of W. Gautschi
    <https://doi.org/10.1093/oso/9780198506720.001.0001>`_.

    Attributes
    ----------
    nodeType : str
        The type of node distribution
    quadType : str
        The quadrature type
    """
    def __init__(self, nodeType='LEGENDRE', quadType='LOBATTO'):
        """
        Parameters
        ----------
        nodeType : str, optional
            The type of node distribution, can be

            - EQUID : equidistant nodes
            - LEGENDRE : node distribution from Legendre polynomials
            - CHEBY-1 : node distribution from Chebychev polynomials (1st kind)
            - CHEBY-2 : node distribution from Chebychev polynomials (2nd kind)
            - CHEBY-3 : node distribution from Chebychev polynomials (3rd kind)
            - CHEBY-4 : node distribution from Chebychev polynomials (4th kind)

            The default is 'LEGENDRE'.

        quadType : str, optional
            The quadrature type, can be

            - GAUSS : inner point only, no node at boundary
            - RADAU-LEFT : only left boundary as node
            - RADAU-RIGHT : only right boundary as node
            - LOBATTO : left and right boundary as node

            The default is 'LOBATTO'.
        """

        # Check argument validity
        for arg, vals in zip(['nodeType', 'quadType'], [NODE_TYPES, QUAD_TYPES]):
            val = eval(arg)
            if val not in vals:
                raise ValueError(f"{arg}='{val}' not implemented, must be in {vals}")

        # Store attributes
        self.nodeType = nodeType
        self.quadType = quadType


    def getNodes(self, nNodes):
        """
        Computes a given number of quadrature nodes.

        Parameters
        ----------
        nNodes : int
            Number of nodes to compute.

        Returns
        -------
        nodes : np.1darray
            Nodes located in [-1, 1], in increasing order.
        """
        # Check number of nodes
        if self.quadType in ['LOBATTO', 'RADAU-LEFT'] and nNodes < 2:
            raise ValueError(
                f"nNodes must be larger than 2 for {self.quadType},"
                f" but got {nNodes}")
        elif nNodes < 1:
            raise ValueError("you surely want at least one node ;)")

        # Equidistant nodes
        if self.nodeType == 'EQUID':
            if self.quadType == 'GAUSS':
                return np.linspace(-1, 1, num=nNodes + 2)[1:-1]
            elif self.quadType == 'LOBATTO':
                return np.linspace(-1, 1, num=nNodes)
            elif self.quadType == 'RADAU-RIGHT':
                return np.linspace(-1, 1, num=nNodes + 1)[1:]
            elif self.quadType == 'RADAU-LEFT':
                return np.linspace(-1, 1, num=nNodes + 1)[:-1]

        # Quadrature nodes linked to orthogonal polynomials
        alpha, beta = self.getTridiagCoefficients(nNodes)
        nodes = eigh_tridiagonal(alpha, np.sqrt(beta[1:]))[0]
        nodes.sort()

        # Force overwrite boundary values if necessary (better precision)
        if self.quadType in ["LOBATTO", "RADAU-RIGHT"]:
            nodes[-1] = 1
        if self.quadType in ["LOBATTO", "RADAU-LEFT"]:
            nodes[0] = -1

        return nodes


    def getOrthogPolyCoefficients(self, num_coeff):
        """
        Produces a given number of analytic three-term recurrence coefficients.

        Parameters
        ----------
        num_coeff : int
            Number of coefficients to compute.

        Returns
        -------
        alpha : np.1darray
            The alpha coefficients of the three-term recurrence.
        beta : np.1darray
            The beta coefficients of the three-term recurrence.
        """
        if self.nodeType == 'LEGENDRE':
            k = np.arange(num_coeff, dtype=float)
            alpha = 0 * k
            beta = k**2 / (4 * k**2 - 1)
            beta[0] = 2
        elif self.nodeType == 'CHEBY-1':
            alpha = np.zeros(num_coeff)
            beta = np.full(num_coeff, 0.25)
            beta[0] = np.pi
            if num_coeff > 1:
                beta[1] = 0.5
        elif self.nodeType == 'CHEBY-2':
            alpha = np.zeros(num_coeff)
            beta = np.full(num_coeff, 0.25)
            beta[0] = np.pi / 2
        elif self.nodeType == 'CHEBY-3':
            alpha = np.zeros(num_coeff)
            alpha[0] = 0.5
            beta = np.full(num_coeff, 0.25)
            beta[0] = np.pi
        elif self.nodeType == 'CHEBY-4':
            alpha = np.zeros(num_coeff)
            alpha[0] = -0.5
            beta = np.full(num_coeff, 0.25)
            beta[0] = np.pi
        return alpha, beta


    def evalOrthogPoly(self, t, alpha, beta):
        """
        Evaluates the two higher order orthogonal polynomials corresponding
        to the given (alpha,beta) coefficients.

        Parameters
        ----------
        t : float or np.1darray
            The point where to evaluate the orthogonal polynomials.
        alpha : np.1darray
            The alpha coefficients of the three-term recurrence.
        beta : np.1darray
            The beta coefficients of the three-term recurrence.

        Returns
        -------
        pi[0] : float or np.1darray
            The second higher order orthogonal polynomial evaluation.
        pi[1] : float or np.1darray
            The higher oder orthogonal polynomial evaluation.
        """
        t = np.asarray(t, dtype=float)
        pi = np.array([np.zeros_like(t) for i in range(3)])
        pi[1:] += 1
        for alpha_j, beta_j in zip(alpha, beta):
            pi[2] *= t - alpha_j
            pi[0] *= beta_j
            pi[2] -= pi[0]
            pi[0] = pi[1]
            pi[1] = pi[2]
        return pi[0], pi[1]


    def getTridiagCoefficients(self, nNodes):
        """
        Computes recurrence coefficients for the tridiagonal Jacobian matrix,
        taking into account the quadrature type.

        Parameters
        ----------
        nNodes : int
            Number of nodes that should be computed from those coefficients.

        Returns
        -------
        alpha : np.1darray
            The modified alpha coefficients of the three-term recurrence.
        beta : np.1darray
            The modified beta coefficients of the three-term recurrence.
        """
        # Coefficients for Gauss quadrature type
        alpha, beta = self.getOrthogPolyCoefficients(nNodes)

        # If not Gauss quadrature type, modify the alpha/beta coefficients
        if self.quadType.startswith('RADAU'):
            b = -1.0 if self.quadType.endswith('LEFT') else 1.0
            b1, b2 = self.evalOrthogPoly(b, alpha[:-1], beta[:-1])[:2]
            alpha[-1] = b - beta[-1] * b1 / b2
        elif self.quadType == 'LOBATTO':
            a, b = -1.0, 1.0
            a2, a1 = self.evalOrthogPoly(a, alpha[:-1], beta[:-1])[:2]
            b2, b1 = self.evalOrthogPoly(b, alpha[:-1], beta[:-1])[:2]
            alpha[-1], beta[-1] = np.linalg.solve([[a1, a2], [b1, b2]], [a * a1, b * b1])
        return alpha, beta



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base module for Q coefficients generation
"""
import numpy as np

from qmat.utils import checkOverriding, storeClass, importAll
from qmat.lagrange import LagrangeApproximation

class QGenerator(object):

    @classmethod
    def getInstance(cls):
        try:
            return cls()
        except TypeError:
            return cls(**cls.DEFAULT_PARAMS)

    @property
    def nodes(self):
        raise NotImplementedError("mouahahah")

    @property
    def Q(self):
        raise NotImplementedError("mouahahah")

    @property
    def weights(self):
        raise NotImplementedError("mouahahah")

    @property
    def weightsEmbedded(self):
        """
        These weights can be used to construct a secondary lower order method from the same stages.
        """
        raise NotImplementedError("no embedded weights implemented for {type(self).__name__}")

    @property
    def nNodes(self):
        return self.nodes.size

    @property
    def rightIsNode(self):
        return self.nodes[-1] == 1.

    @property
    def T(self):
        """Transfer matrix from zero-to-nodes to node-to-node"""
        M = self.Q.shape[0]
        T = np.eye(M)
        T[1:,:-1][np.diag_indices(M-1)] = -1
        return T

    @property
    def S(self):
        Q = np.asarray(self.Q)
        M = self.Q.shape[0]
        T = np.eye(M)
        T[1:,:-1][np.diag_indices(M-1)] = -1
        return T @ Q

    @property
    def Tinv(self):
        """Transfer matrix from node-to-node to zero-to-node"""
        M = self.Q.shape[0]
        return np.tri(M)

    @property
    def hCoeffs(self):
        approx = LagrangeApproximation(self.nodes)
        return approx.getInterpolationMatrix([1]).ravel()

    def genCoeffs(self, withS=False, hCoeffs=False, embedded=False):
        out = [self.nodes, self.weights, self.Q]

        if embedded:
            out[1] = np.vstack([out[1], self.weightsEmbedded])
        if withS:
            out.append(self.S)
        if hCoeffs:
            out.append(self.hCoeffs)
        return out

    @property
    def order(self):
        raise NotImplementedError("mouahahah")

    @property
    def orderEmbedded(self):
        return self.order - 1

    def solveDahlquist(self, lam, u0, T, nSteps, useEmbeddedWeights=False):
        nodes, weights, Q = self.nodes, self.weights, self.Q

        if useEmbeddedWeights:
            weights = self.weightsEmbedded

        uNum = np.zeros(nSteps+1, dtype=complex)
        uNum[0] = u0

        dt = T/nSteps
        A = np.eye(nodes.size) - lam*dt*Q
        for i in range(nSteps):
            b = np.ones(nodes.size)*uNum[i]
            uStages = np.linalg.solve(A, b)
            uNum[i+1] = uNum[i] + lam*dt*weights.dot(uStages)

        return uNum

    def errorDahlquist(self, lam, u0, T, nSteps, uNum=None, useEmbeddedWeights=False):
        if uNum is None:
            uNum = self.solveDahlquist(lam, u0, T, nSteps, useEmbeddedWeights=useEmbeddedWeights)
        times = np.linspace(0, T, nSteps+1)
        uExact = u0 * np.exp(lam*times)
        return np.linalg.norm(uNum-uExact, ord=np.inf)


Q_GENERATORS = {}

def register(cls:QGenerator)->QGenerator:
    # Check for correct overriding
    for name in ["nodes", "Q", "weights", "order"]:
        checkOverriding(cls, name)
    # Check that TEST_PARAMS are given and valid if no default constructor
    try:
        cls()
    except TypeError:
        try:
            params = cls.DEFAULT_PARAMS
        except AttributeError:
            raise AttributeError(
                f"{cls.__name__} requires DEFAULT_PARAMS attribute"
                " since it has no default constructor")
        try:
            cls(**params)
        except:
            raise TypeError(
                f"{cls.__name__} could not be instantiated with DEFAULT_PARAMS")
    # Store class (and aliases)
    storeClass(cls, Q_GENERATORS)
    return cls

def genQCoeffs(qType, withS=False, hCoeffs=False, embedded=False, **params):
    try:
        Generator = Q_GENERATORS[qType]
    except KeyError:
        raise ValueError(f"{qType=!r} is not available")
    gen = Generator(**params)
    return gen.genCoeffs(withS, hCoeffs, embedded)


# Import all local submodules
__all__ = importAll(locals(), __path__, __name__, __import__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Utility module
"""
import inspect
import pkgutil
import numpy as np

def checkOverriding(cls, name, isProperty=True):
    method = getattr(cls, name)
    parent = getattr(cls.mro()[-2], name)
    assert method != parent, \
        f"{name} method must be overriden in {cls.__name__} class"
    if isProperty:
        assert type(method) == property, \
            f"{name} method must be a property in {cls.__name__} class"
    else:
        pass
        # TODO : check that signatures are the same


def checkGenericConstr(cls):
    sig = inspect.signature(cls.__init__)
    try:
        par = sig.parameters["kwargs"]
        assert par.kind == par.VAR_KEYWORD
    except (KeyError, AssertionError):
        raise AssertionError(f"{cls.__name__} class requires **kwargs in its constructor")


def storeAlias(cls, dico, alias):
    assert alias not in dico, f"{alias} alias already registered in {dico}"
    dico[alias] = cls


def storeClass(cls, dico):
    storeAlias(cls, dico, cls.__name__)
    aliases = getattr(cls, "aliases", None)
    if aliases:
        assert isinstance(aliases, list), \
            f"aliases must be a list in class {cls.__name__}"
        for alias in aliases:
            storeAlias(cls, dico, alias)


def importAll(localVars, path, name, _import):
    """The magic function"""
    _all = [var for var in localVars.keys() if not var.startswith('__')]
    for _, moduleName, _ in pkgutil.walk_packages(path):
        _all.append(moduleName)
        _import(name+'.'+moduleName)


def getClasses(dico, module=None):
    classes = {}
    if module is None:
        check = lambda cls: True
    else:
        check = lambda cls: cls.__module__.endswith("."+module)
    for key, cls in dico.items():
        if cls not in classes.values() and check(cls):
            classes[key] = cls
    return classes


def numericalOrder(nSteps, err):
    """
    Utility function to compute numerical order from error and nSteps vectors

    Parameters
    ----------
    nSteps : np.1darray or list
        Different number of steps to compute the error.
    err : np.1darray
        Diffenrent error values associated to the number of steps.

    Returns
    -------
    beta : float
        Order coefficient computed through linear regression.
    rmse : float
        The root mean square error of the linear regression.
    """
    nSteps = np.asarray(nSteps)
    x, y = np.log10(1/nSteps), np.log10(err)

    # Compute regression coefficients and rmse
    xMean = x.mean()
    yMean = y.mean()
    sX = ((x-xMean)**2).sum()
    sXY = ((x-xMean)*(y-yMean)).sum()

    beta = sXY/sX
    alpha = yMean - beta*xMean

    yHat = alpha + beta*x
    rmse = ((y-yHat)**2).sum()**0.5
    rmse /= x.size**0.5

    return beta, rmse



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Submodule for QDelta coefficients based on time-stepping scheme
"""
import numpy as np

from qmat.qdelta import QDeltaGenerator, register

class TimeStepping(QDeltaGenerator):

    def __init__(self, nodes, **kwargs):
        self.nodes = np.asarray(nodes)
        deltas = nodes.copy()
        deltas[1:] = np.ediff1d(nodes)
        self.deltas = deltas
        M = self.nodes.size
        self.QDelta = np.zeros((M, M), dtype=float)


@register
class BE(TimeStepping):
    """Approximation based on Backward Euler steps between the nodes"""
    aliases = ["IE"]

    def getQDelta(self, k=None):
        QDelta, M, deltas = self.QDelta, self.nodes.size, self.deltas
        for i in range(M):
            QDelta[i:, :M-i] += np.diag(deltas[:M-i])
        return self.storeAndReturn(QDelta)


@register
class FE(TimeStepping):
    """Approximation based on Forward Euler steps between the nodes"""
    aliases = ["EE"]

    def getQDelta(self, k=None):
        QDelta, M, deltas = self.QDelta, self.nodes.size, self.deltas
        for i in range(1, M):
            QDelta[i:, :M-i] += np.diag(deltas[1:M-i+1])
        return self.storeAndReturn(QDelta)

    @property
    def dTau(self):
        return self.nodes*0 + self.deltas[0]

@register
class TRAP(TimeStepping):
    """Approximation based on Trapezoidal Rule between the nodes"""
    aliases = ["CN"]

    def getQDelta(self, k=None):
        QDelta, M, deltas = self.QDelta, self.nodes.size, self.deltas
        for i in range(0, M):
            QDelta[i:, :M-i] += np.diag(deltas[:M-i])
        for i in range(1, M):
            QDelta[i:, :M-i] += np.diag(deltas[1:M-i+1])
        QDelta /= 2.0
        return QDelta

    @property
    def dTau(self):
        return self.nodes*0 + self.deltas[0]/2.0


@register
class BEPAR(TimeStepping):
    """Approximation based on parallel Backward Euler steps from zero to nodes"""
    aliases = ["IEpar"]

    def getQDelta(self, k=None):
        self.QDelta[:] = np.diag(self.nodes)
        return self.QDelta


@register
class TRAPAR(TimeStepping):
    """Approximation based on parallel Trapezoidal Rule from zero to nodes"""

    def getQDelta(self, k=None):
        self.QDelta[:] = np.diag(self.nodes/2)
        return self.QDelta

    @property
    def dTau(self):
        return self.nodes/2.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base module for QDelta coefficients generation
"""
import numpy as np

from qmat.utils import checkOverriding, storeClass, importAll, checkGenericConstr

class QDeltaGenerator(object):

    def __init__(self, Q, **kwargs):
        self.Q = np.asarray(Q, dtype=float)
        self.QDelta = np.zeros_like(self.Q)

    def storeAndReturn(self, QDelta):
        np.copyto(self.QDelta, QDelta)
        return self.QDelta

    def getQDelta(self, k=None):
        raise NotImplementedError("mouahahah")

    @property
    def dTau(self):
        return self.QDelta[0]*0

    def genCoeffs(self, k=None, dTau=False):
        if isinstance(k, list):
            out = [np.array([self.getQDelta(_k) for _k in k])]
        else:
            out = [self.getQDelta(k)]
        if dTau:
            out += [self.dTau]
        return out if len(out) > 1 else out[0]

QDELTA_GENERATORS = {}

def register(cls:QDeltaGenerator)->QDeltaGenerator:
    checkGenericConstr(cls)
    checkOverriding(cls, "getQDelta", isProperty=False)
    storeClass(cls, QDELTA_GENERATORS)
    return cls

def genQDeltaCoeffs(qDeltaType, nSweeps=None, dTau=False, **params):

    # Check arguments
    if isinstance(qDeltaType, str):
        if nSweeps is None:
            pass  # only one QDelta matrix, default approach
        elif isinstance(nSweeps, int) and nSweeps > 0:
            qDeltaType = [qDeltaType]  # more sweeps of the same QDelta matrix
        else:
            raise ValueError(f"bad value of nSweep {nSweeps}")
    elif isinstance(qDeltaType, list):
        assert len(qDeltaType) > 0, "need at least one qDeltaType in the list"
        if nSweeps is None:
            nSweeps = len(qDeltaType)  # number of sweeps given in the list
        elif isinstance(nSweeps, int) and nSweeps > 0:
            # complete with additional sweeps
            assert nSweeps >= len(qDeltaType), \
                f"nSweeps ({nSweeps} lower than list length for qDeltaType ({qDeltaType})"
            qDeltaType += [qDeltaType[-1]]*(nSweeps-len(qDeltaType))
        else:
            raise ValueError(f"bad value of nSweep {nSweeps}")
    else:
        raise ValueError(f"bad value of qDeltaType {qDeltaType}")

    if nSweeps is None:  # Single matrix return

        try:
            Generator = QDELTA_GENERATORS[qDeltaType]
        except KeyError:
            raise ValueError(f"qDeltaType={qDeltaType} is not available")

        gen = Generator(**params)
        return gen.genCoeffs(dTau=dTau)

    else:  # Multiple matrices return
        try:
            Generators = [QDELTA_GENERATORS[qDT] for qDT in qDeltaType]
        except KeyError:
            raise ValueError(f"qDeltaType={qDeltaType} is not available")

        if len(qDeltaType) == 1:  # Single QDelta generator
            gen = Generators[0](**params)
            return gen.genCoeffs(k=[k+1 for k in range(nSweeps)], dTau=dTau)

        else:  # Multiple QDelta generators
            gens = [Gen(**params) for Gen in Generators]
            out = [np.array(
                [gen.getQDelta(k+1) for k, gen in enumerate(gens)]
                )]
            if dTau:
                out += [gens[0].dTau]

    return out if len(out) > 1 else out[0]


# Import all local submodules
__all__ = importAll(locals(), __path__, __name__, __import__)



import pytest
import numpy as np

from qmat.qdelta import QDELTA_GENERATORS, genQDeltaCoeffs
from qmat.utils import getClasses

# Test base functionnality only on algebraic QDelta-generators
GENERATORS = getClasses(QDELTA_GENERATORS, module="algebraic")


@pytest.mark.parametrize("nNodes", [2, 3, 4])
@pytest.mark.parametrize("name", GENERATORS.keys())
def testGeneration(name, nNodes):

    Q = np.random.rand(nNodes, nNodes)
    gen = GENERATORS[name](Q=Q)

    QD1 = gen.genCoeffs()
    assert type(QD1) == np.ndarray, \
        f"QDelta for {name} is not np.ndarray but {type(QD1)}"
    assert QD1.shape == (nNodes, nNodes), \
        f"QDelta for {name} has unconsistent shape : {QD1.shape}"
    if name != "Exact":
        assert np.allclose(np.tril(QD1), QD1), \
            f"QDelta for {name} is not lower triangular"

    QD2 = genQDeltaCoeffs(name, Q=Q)
    assert QD1.shape == QD2.shape, \
        f"OOP QDelta and PP QDelta have not the same shape for {name}"
    assert np.allclose(QD1, QD2), \
        f"OOP QDelta and PP QDelta are not equals for {name}"

    _, dTau1 = gen.genCoeffs(dTau=True)
    assert type(dTau1) == np.ndarray, \
        f"dTau for {name} is not np.ndarray but {type(dTau1)}"
    assert dTau1.ndim == 1, \
        f"dTau for {name} is not 1D : {dTau1}"
    assert dTau1.size == nNodes, \
        f"dTau for {name} has not the correct size : {dTau1}"

    _, dTau2 = genQDeltaCoeffs(name, Q=Q, dTau=True)
    assert np.allclose(dTau1, dTau2), \
        f"OOP dTau and PP dTau are not equals for {name}"


nNodes = 4
@pytest.mark.parametrize("nSweeps", [1, 2, 3])
@pytest.mark.parametrize("name", GENERATORS.keys())
def testSingleGenerationMultipleSweeps(name, nSweeps):
    Q = np.random.rand(nNodes, nNodes)
    gen = GENERATORS[name](Q=Q)

    QD1 = gen.genCoeffs(k=[k+1 for k in range(nSweeps)])
    assert QD1.shape == (nSweeps, nNodes, nNodes), \
        f"QDelta for {name} has unconsistent shape : {QD1.shape}"

    QD2 = genQDeltaCoeffs(name, nSweeps=nSweeps, Q=Q)
    assert QD1.shape == QD2.shape, \
        f"OOP QDelta and PP QDelta have not the same shape for {name}"
    assert np.allclose(QD1, QD2), \
        f"OOP QDelta and PP QDelta are not equals for {name}"

    _, dTau1 = gen.genCoeffs(k=[k+1 for k in range(nSweeps)], dTau=True)
    _, dTau2 = genQDeltaCoeffs(name, nSweeps=nSweeps, dTau=True, Q=Q)
    assert dTau1.shape == dTau2.shape, \
        f"OOP dTau and PP dTau have not the same shape for {name}"
    assert np.allclose(dTau1, dTau2), \
        f"OOP dTau and PP dTau are not equals for {name}"


@pytest.mark.parametrize("nSweeps", [1, 2, 3, 4])
def testMultipleGenerationMultipleSweeps(nSweeps):
    Q = np.random.rand(nNodes, nNodes)

    names = list(GENERATORS.keys())[:nSweeps]
    generators = [GENERATORS[name](Q=Q) for name in names]

    QD1 = np.array([gen.genCoeffs() for gen in generators])
    QD2 = genQDeltaCoeffs(names, Q=Q)
    assert QD1.shape == QD2.shape, \
        f"OOP QDelta and PP QDelta have not the same shape with K={nSweeps}"
    assert np.allclose(QD1, QD2), \
        f"OOP QDelta and PP QDelta are not equals with K={nSweeps}"

    QD3 = genQDeltaCoeffs(names, nSweeps=nSweeps+1, Q=Q)
    assert QD3.shape[0] == QD1.shape[0]+1 and QD3.shape[1:] == QD1.shape[1:],\
        f"sweep extension produces inconsistent shapes for K={nSweeps}"
    assert np.allclose(QD3[-1], QD3[-2]), \
        f"sweep extension don't re-use the same matrix for K={nSweeps}"

    _, dTau1 = genQDeltaCoeffs(names, dTau=True, Q=Q)
    _, dTau2 = genQDeltaCoeffs(names, nSweeps=nSweeps+1, dTau=True, Q=Q)
    assert dTau1.shape == dTau2.shape, \
        f"sweep extension produces inconsistent dTau shapes for K={nSweeps}"
    assert np.allclose(dTau1, dTau2), \
        f"sweep extension don't re-use the same dTau for K={nSweeps}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from qmat.qcoeff import genQCoeffs, Q_GENERATORS
from qmat.qdelta import genQDeltaCoeffs, QDELTA_GENERATORS

__all__ = [
    "genQCoeffs", "genQDeltaCoeffs",
    "Q_GENERATORS", "QDELTA_GENERATORS"]



#!/usr/bin/env python
from qmat import genQCoeffs

def main():
    nodes, weights, Q = genQCoeffs("coll", nNodes=3, nodeType="LEGENDRE", quadType="GAUSS")
    print("node :", nodes)
    print("weights :", weights)
    print("Q :")
    print(Q)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module with simple usage SDC type time-stepping solvers
"""
import numpy as np


def solveDahlquistSDC(lam, u0, T, nSteps, nSweeps, Q, QDelta,
                      weights=None):
    uNum = np.zeros(nSteps+1, dtype=complex)
    uNum[0] = u0

    nNodes = Q.shape[0]

    dt = T/nSteps
    P = np.eye(nNodes) - lam*dt*QDelta
    for i in range(nSteps):

        uNodes = np.ones(nNodes)*uNum[i]

        for k in range(nSweeps):
            b = uNum[i] + lam*dt*(Q-QDelta) @ uNodes
            uNodes = np.linalg.solve(P, b)

        if weights is None:
            uNum[i+1] = uNodes[-1]
        else:
            uNum[i+1] = uNum[i] + lam*dt*weights.dot(uNodes)

    return uNum


def errorDahlquistSDC(lam, u0, T, nSteps, nSweeps, Q, QDelta,
                      weights=None, uNum=None):
    if uNum is None:
        uNum = solveDahlquistSDC(
            lam, u0, T, nSteps, nSweeps, Q, QDelta,
            weights=weights)

    times = np.linspace(0, T, nSteps+1)
    uExact = u0 * np.exp(lam*times)
    return np.linalg.norm(uNum-uExact, ord=np.inf)


def getOrderSDC(coll, nSweeps, qDelta, prolongation):
    # TODO : extend with additional results from
    # https://gitlab.inria.fr/sweet/sweet/-/blob/main/mule_local/python/sdc/qmatrix.py#L596

    nNodes, nodeType, quadType = coll.nodes.size, coll.nodeType, coll.quadType

    # Maximum order from the collocation problem
    maxOrder = coll.order

    # Order of SDC
    order = 0
    if nSweeps > 0:
        # first sweep
        if qDelta in ['TRAPAR', 'TRAP']:
            order += 2
        else:
            order += 1
        # rest of sweeps
        order += nSweeps-1
    # take into account prolongation
    if prolongation == "QUADRATURE":
        order += 1

    order = min(maxOrder, order)

    # Edge cases with bonus order (or malus some times ...)
    # TODO: couple with the Butcher theory from Joscha to retrieve this theoretically ...
    if prolongation == "QUADRATURE":  # COPY initialization
        if qDelta == "TRAP":
            if nSweeps == 1 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "CHEBY-2" and quadType == "LOBATTO":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "CHEBY-4" and quadType in ["RADAU-RIGHT", "LOBATTO"]:
                order += 1
        if qDelta == "TRAPAR":
            if nSweeps == 2 and nNodes == 3 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 2 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
        if qDelta == "BE":
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "FE":
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-1" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-3" and quadType in ["RADAU-LEFT", "LOBATTO"]:
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "BEPAR":
            if nSweeps == 3 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType in ["CHEBY-1", "CHEBY-2"] and quadType in ["RADAU-LEFT", "RADAU-RIGHT"]:
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType in ["CHEBY-3", "CHEBY-4"]:
                order += 1
        if qDelta == "TRAPAR":
            if nSweeps == 3 and nNodes == 4 and nodeType == "EQUID":
                order += 1
            if nSweeps == 3 and nNodes == 4 and nodeType in ["CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"]:
                order += 1

    if prolongation == "LASTNODE":
        if qDelta == "BE":
            if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "FE":
            if nSweeps == 3 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "TRAP":
            if nSweeps == 2 and nNodes == 3 and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2"] and quadType == "LOBATTO":
                order += 1
            if nSweeps == 2 and nNodes == 3 and nodeType == "LEGENDRE":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "EQUID" and quadType == "LOBATTO":
                order += 2
            if nSweeps == 2 and nNodes == 4 and nodeType == "EQUID" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType in ["LEGENDRE", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"]:
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "LEGENDRE" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 4 and nodeType == "LEGENDRE" and quadType == "LOBATTO":
                order += 1
            if nSweeps == 4 and nNodes == 4 and nodeType == "LEGENDRE" and quadType == "RADAU-RIGHT":
                order += 2
        if qDelta == "BEPAR":
            if nSweeps == 3 and nNodes == 2 and nodeType == "CHEBY-2" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 3 and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"] and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 3 and nodeType in ["CHEBY-3", "CHEBY-4"] and quadType == "LOBATTO":
                order += 1
        if qDelta == "TRAPAR":
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 4 and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"]:
                order += 1


    return order



import pytest
import numpy as np

from qmat.lagrange import LagrangeApproximation

nNodeTests = [2, 3, 4, 5, 6, 7, 8]

def referenceWeights(pType, n):
    j = np.arange(n)
    if pType == 'CHEBY-1':
        j = j[-1::-1]
        points = np.cos((2*j+1)*np.pi/(2*n))
        weights = np.sin((2*j+1)*np.pi/(2*n)) * (-1)**j
    elif pType == 'CHEBY-2':
        j = j[-1::-1]
        points = np.cos(j*np.pi/(n-1))
        weights = np.array([0.5]+[1]*(n-2)+[0.5]) * (-1)**j
    weights /= np.max(np.abs(weights))
    return points, weights


@pytest.mark.parametrize("weightComputation", ["AUTO", "FAST", "STABLE", "CHEBFUN"])
@pytest.mark.parametrize("pType", ["CHEBY-1", "CHEBY-2"])
def testWeights(pType, weightComputation):
    for n in nNodeTests:
        points, weights = referenceWeights(pType, n)
        approx = LagrangeApproximation(points, weightComputation=weightComputation)
        assert np.allclose(approx.weights, weights), f"discrepancy with reference weights for n={n}"


@pytest.mark.parametrize("weightComputation", ["AUTO", "CHEBFUN"])
@pytest.mark.parametrize("pType", ["CHEBY-1", "CHEBY-2"])
def testAsymptoticWeights(pType, weightComputation):
    points, weights = referenceWeights(pType, 8000)
    approx = LagrangeApproximation(points, weightComputation=weightComputation)
    assert np.allclose(approx.weights, weights)


@pytest.mark.parametrize("scaleRef", ["MAX", "ZERO"])
@pytest.mark.parametrize("pType", ["CHEBY-1", "CHEBY-2"])
def testAsymptoticWeightsSTABLE(pType, scaleRef):
    points, weights = referenceWeights(pType, 10000)
    approx = LagrangeApproximation(points, weightComputation="STABLE", scaleRef=scaleRef)
    assert np.allclose(approx.weights, weights)


@pytest.mark.parametrize("weightComputation", ["AUTO", "FAST", "STABLE", "CHEBFUN"])
@pytest.mark.parametrize("nNodes", nNodeTests)
def testInterpolation(nNodes, weightComputation):
    nodes = np.sort(np.random.rand(nNodes))
    approx = LagrangeApproximation(nodes, weightComputation=weightComputation)
    
    times = np.random.rand(nNodes*2)
    P = approx.getInterpolationMatrix(times)

    polyCoeffs = np.random.rand(nNodes)
    polyNodes = np.polyval(polyCoeffs, nodes)
    polyTimes = np.polyval(polyCoeffs, times)
    assert np.allclose(polyTimes, P @ polyNodes)
    

@pytest.mark.parametrize("numQuad", ["LEGENDRE_NUMPY", "LEGENDRE_SCIPY", "FEJER"])
@pytest.mark.parametrize("weightComputation", ["AUTO", "FAST", "STABLE", "CHEBFUN"])
@pytest.mark.parametrize("nNodes", nNodeTests)
def testIntegration(nNodes, weightComputation, numQuad):
    nodes = np.sort(np.random.rand(nNodes))
    approx = LagrangeApproximation(nodes, weightComputation=weightComputation)

    times = np.random.rand(nNodes*2)
    P = approx.getIntegrationMatrix([(0, t) for t in times], numQuad=numQuad)

    polyCoeffs = np.random.rand(nNodes)
    polyNodes = np.polyval(polyCoeffs, nodes)
    polyInteg = np.polyval(np.polyint(polyCoeffs), times) - np.polyval(np.polyint(polyCoeffs), 0)

    assert np.allclose(polyInteg, P @ polyNodes)